open Util

open Format

type desc = {
  id: int;
  format: formatter -> unit;
}

type t = {
  desc: desc;
  store: unit -> unit;
}

module Type = struct
  type 'a ty = {
    inj: 'a -> t;
    proj: t -> 'a;
    check: t -> bool;
  }
end

open Type

let count = ref 0

let new_type format =
  let id = !count in
  count := succ id;
  let cell = ref None in
  let desc =
    { id = id; format = (fun fmt -> format fmt (Option.get !cell)) } in
  let inj x = { desc = desc; store = (fun () -> cell := Some x) } in
  let check x = x.desc.id = id in
  let proj x =
    if not (check x) then
      failwith "Invalid projection";
    x.store ();
    Option.get !cell in
  { inj = inj; proj = proj; check = check }

let format fmt d =
  d.store ();
  d.desc.format fmt
