type ('a, 'b) point = ('a, 'b) node ref
and ('a, 'b) node =
  | Link of ('a, 'b) point
  | Desc of ('a, 'b) desc
and ('a, 'b) desc =
  | Fresh of 'a
  | Value of 'b

let of_desc d = ref (Desc d)

let fresh p = of_desc (Fresh p)

let of_value v = of_desc (Value v)

let rec find v =
  match !v with
    | Link v' ->
        let (r, _) as result = find v' in
        if not (v' == r) then
          v := Link r;
        result
    | Desc d -> (v, d)

exception Fail

let rec union unify pos v0 v1 =
  let (r0, d0) = find v0 and (r1, d1) = find v1 in
  if r0 == r1 then
    Warn.redundent_constraint pos
  else
    match d0, d1 with
      | Fresh _, _ -> r0 := Link r1
      | _, Fresh _ -> r1 := Link r0
      | Value v0, Value v1 ->
          begin
            r1 := Link r0;
            try unify v0 v1
            with Fail ->
              r1 := Desc d1;
              raise Fail
          end
